// @ myth - client - input.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "client/input.h"
#include "common/error.h"
#include "common/parse.h"
#include "client/macro.h"
#include "client/output.h"
#include "common/ipc.h"
#include "common/command.h"
#include "client/question.h"
#include "common/format.h"
#include "client/global.h"

// @@ MACRO macro

static MACRO macro(format("%s/%s/%s", HomeDirectory, LibraryDirectory, "KEY"));

// @@ INPUT

// @@@ void INPUT::clear()
//
// Remove the question.

void INPUT::clear()
{
    if(question != NULL)
    {
	delete question;
	question = NULL;
    }
    argument = 0L;
}

// @@@ QUESTION* INPUT::current()
//
// Return the subquestion being answered now.

QUESTION* INPUT::current()
{
    if(question == NULL)
	return NULL;
    else if(question->done())
	return question->current();
    else
	return question;
}

// @@@ void INPUT::ask(CHANNEL* channel)
//
// Read a request from the server.

void INPUT::ask(CHANNEL* channel)
{
    if(question != NULL)
	bug.error("INPUT::ask", "Unanswered question");
    // Forget about the old paged message
    page.clear();

    // All must contain a prompt and the type of request
    prompt_type prompt = channel->GetChar();
    request_type request = channel->GetChar();
    // Some request take additional arguments
    inventory_type bunch;
    symbol_type symbol;

    switch(request)
    {
    case REQUEST_OBJECT:
	bunch = channel->GetChar();
	question = new QUESTION_OBJECT(prompt, bunch);
	return;
    case REQUEST_POSITION:
	question = new QUESTION_POSITION(prompt);
	return;
    case REQUEST_STRING:
	question = new QUESTION_STRING(prompt);
	return;
    case REQUEST_DIRECTION:
	question = new QUESTION_DIRECTION(prompt);
	return;
    case REQUEST_SYMBOL:
	bunch = channel->GetChar();
	question = new QUESTION_SYMBOL(prompt, bunch);
	return;
    case REQUEST_CONFIRMATION:
	question = new QUESTION_CONFIRMATION(prompt);
	return;
    case REQUEST_ARGUMENT:
	question = new QUESTION_ARGUMENT(prompt);
	return;
    case REQUEST_LETTER:
	bunch = channel->GetChar();
	symbol = channel->GetChar();
	question = new QUESTION_LETTER(prompt, bunch, symbol);
	return;
    case REQUEST_NAME:
	question = new QUESTION_NAME(prompt);
	return;
    case REQUEST_COMMAND:
	question = new QUESTION_COMMAND(prompt);
	return;
    case REQUEST_IDLE:
    case REQUEST_END:
    case REQUEST_ERROR:
    default:
	bug.error("INPUT::read", format("Unknown request %d", request));
    }
}

// @@@ void INPUT::read(char c, bool paging)
//
// Use the character "c", to answer the current request.
// If "paging" is TRUE, it will eximane if "c" is used to shift page.

void INPUT::read(char c, bool paging)
{
    do_update = TRUE;
    do_escape = FALSE;
    do_redraw = FALSE;
    do_help = FALSE;
    do_page = FALSE;
    error_value = NULL;
    macro.put(c);

    if(current() == NULL)
	bug.error("INPUT::read", "NULL current()");
    else if((paging || page.length() > 0) &&
	    macro.check(KEY_PAGE))
	do_page = TRUE;
    else if(macro.check(KEY_REDRAW))
	do_redraw = TRUE;
    else if(macro.check(KEY_ESCAPE))
    {
	clear();
	do_escape = TRUE;
    }
    else if(macro.check(KEY_HELP))
	do_help = TRUE;
    else if(macro.argument())
	argument = argument * 10 + macro.value();
    else
    {
	while(macro.ready() && current() != NULL && error_value == NULL)
	    error_value = current()->read(&macro, argument);

	if(error_value != NULL)
	    while(macro.ready())
		(void) macro.key();
    }
}

// @@@ void INPUT::update(OUTPUT* output)
//
// Update the screen, reflecting the current status of the input.

void INPUT::update(OUTPUT* output)
{
    // Always check for redraw first
    if(current() != NULL && do_redraw && do_update)
	output->dungeon.redraw();
    // News, status and inventory
    output->news.update(do_update ? FORCE : CHECK);
    output->status.set(STATUS_NEWS, output->news.status());
    output->status.update();
    output->inventory.update();
    // Set covering window (help, inventory, page)
    if(current() == NULL)
	return;
    else if(!do_update)
	/* Do nothing */;
    else if(do_help)
	current()->help(output);
    else if(output->inventory.check())
	output->cover(output->inventory.cover());
    else if(output->cover.active())
    {
	if(do_page)
	    output->cover();
    }
    else if(page.length() > 0)
    {
	page.reset();
	output->cover(&page, /* retry = */ TRUE);
    }
    // Set dungeon, prompt and cursor
    current()->update(output, argument, error_value);
    // Screen has been updated
    do_update = FALSE;
}    

// @@@ bool INPUT::check(OUTPUT* output)
//
// If a local command is ready from input, then execute it and return TRUE.

bool INPUT::check(OUTPUT* output)
{
    // Get rid of the old menu.
    if(!global.sticky())
	output->cover();

    if(do_escape)
	return FALSE;

    if(question == NULL)
    {
	bug.error("INPUT::check", "NULL question");
	return FALSE;
    }
    return question->check(output, argument);
}

// @@@ void INPUT::answer(CHANNEL* channel)

void INPUT::answer(CHANNEL* channel)
{
    if(do_escape)
	channel->PutChar(ANSWER_ESCAPE);
    else if(question == NULL)
	bug.error("INPUT::answer", "NULL question");
    else
    {
	if(argument != 0L)
	{
	    channel->PutChar(ANSWER_ARGUMENT);
	    channel->PutLong(argument);
	}
	else
	    channel->PutChar(ANSWER);
	
	question->answer(channel);
	clear();
    }
    channel->flush();
}

// @@@ void INPUT::EndGame()

void INPUT::EndGame()
{
    clear();
    question = new QUESTION_GAME_OVER;
}

// @@@ INPUT::INPUT()

INPUT::INPUT()
{
    do_escape = FALSE;
    do_redraw = FALSE;
    do_help = FALSE;
    do_page = FALSE;
    do_update = TRUE;
    error_value = NULL;
    question = NULL;
    argument = 0L;
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
