// @ myth - server - monster.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/rnd.h"
#include "server/monster.h"
#include "common/parse.h"
#include "server/level.h"
#include "common/error.h"
#include "server/player.h"
#include "server/hiscore.h"
#include "common/format.h"

// @@ MONSTER

// @@@ OBJECT* MONSTER::weapon()

OBJECT* MONSTER::weapon()
{
    OBJECT* weapon = pack.weapon();

    if(weapon == NULL)
	return body->weapon();
    else
	return weapon;
}

// @@@ char MONSTER::look()

char MONSTER::look()
{
    if(pack.look(FALSE) == SYMBOL_ERROR)
    {
	if(!CheckAbility(ABILITY_INVISIBLE))
	    return body->look();

	if(pack.size() > body->size())
	{
	    if(pack.look(TRUE) == SYMBOL_ERROR)
		return body->look();

	    return pack.look(TRUE);
	}
	return body->look();
    }
    if(CheckAbility(ABILITY_INVISIBLE))
	return pack.look(FALSE);

    if(pack.size() > body->size())
	return pack.look(FALSE);

    return body->look();
}

// @@@ STRING MONSTER::print(print_flag flag, long count, PLAYER* player)

STRING MONSTER::print(print_flag flag, long count, PLAYER* player)
{
    STRING name = OBJECT::print(flag, count, player);
    
    if((flag & PRINT_VALUE) && body->tainted())
    {
	if(IsUnique())
	    return format("%s (tainted)", name());

	return format("tainted %s", name());
    }
    return name;
}

// @@@ bool MONSTER::visible()

bool MONSTER::visible()
{
    if(!CheckAbility(ABILITY_INVISIBLE))
	return TRUE;
    else if(pack.look(FALSE) != SYMBOL_ERROR)
	return TRUE;
    else
	return FALSE;
}

// @@@ void MONSTER::NewLevel(level_type depth, char where)

void MONSTER::NewLevel(level_type depth, char where)
{
    DungeonLevel = depth;
    NewLocation = where;

    if(depth == LEVEL_EXIT)
    {
	SPECIFY_TYPE specify(AMULET_OF_YENDOR);

	if(pack.contain(&specify))
	{
	    message("The amulet is worth 5000 zorkmids...");
	    message("That's all. You can return to work now.");
	    MyScore += 5000;
	    MyScore *= 2;
	    rip = STRONG("won");
	}
	else
	    rip = STRONG("escaped the dungeon");
    }
}

// @@@ void MONSTER::enter(LEVEL* level)

void MONSTER::enter(LEVEL* level)
{
    DungeonLevel = level->depth();
    // Add it among the others
    level->AddMonster(this);
    // Put it on the bunch
    coor pos = level->locate(NewLocation);
    if(pos == NoWhere)
	move(level, position());
    else
	move(level, pos);
}    

// @@@ bool MONSTER::IsConfused()

bool MONSTER::IsConfused()
{
    return rnd.chance(GetValue(ABILITY_CONFUSED), 25);
}

// @@@ void MONSTER::leave(LEVEL* level)

void MONSTER::leave(LEVEL* level)
{
    if(position() != NoWhere)
    {
	BUNCH* bunch = level->create(position());
	bunch->remove(this);
	level->update(position());
    }
}

// @@@ void MONSTER::teleport(LEVEL* level,

void MONSTER::teleport(LEVEL* level, int length, holy_type holy = NORMAL)
{
    if(position() == NoWhere)
	return;

    // NOT YET IMPLEMENTED! Blessed should allow monster to move directly
    // to their target.

    int CountDown = 100;	// Number of tries

    for(;;)
    {
	coor pos = coor(rnd(-length, length), rnd(-length, length));

	if(CountDown-- < 1)
	    break;
	else if(sqr(pos.x) + sqr(pos.y) > sqr(length))
	    continue;
	else if(holy != CURSED && IsRock(level->rock(pos)))
	    continue;
	else
	{
	    move(level, pos);
	    break;
	}
    }
}

// @@@ void MONSTER::move(LEVEL* level, coor pos)

void MONSTER::move(LEVEL* level, coor pos)
{
    leave(level);
    where = pos;		// Place it at the new pos
    if(pos != NoWhere)		// Put it in a bunch somewhere
    {
	BUNCH* bunch = level->create(pos);
	bunch->add(this);
	switch(level->rock(pos))
	{
	case FLOOR:
	    UseMovePoint(75);
	    break;
	case UNEVEN_FLOOR:
	    UseMovePoint(150);
	    break;
	default:
	    if(CheckAbility(ABILITY_PHASING))
		UseMovePoint(250);
	    else
	    {
		message("You unify with the wall!");
		level->WatchMonster(this, "%s unifies with the wall!");
		kill("was traped in solid rock", level);
	    }
	}
    }
}

// @@@ void MONSTER::kill(STRING why, LEVEL* level)

void MONSTER::kill(STRING why, LEVEL* level)
{
    // Tell the HiScore why it died
    rip = why;
    // Remove the monster...
    if(position() != NoWhere && level != NULL)
    {
	BUNCH* bunch = level->create(position());
	bunch->remove(this);
    }
    // Turn it to a corpse
    body->kill();
    // And add the corpse + inventory
    if(position() != NoWhere && level != NULL)
    {
	BUNCH* bunch = level->create(position());
	bunch->add(this);
	pack.DropDead(bunch);
	pack.update(level, position(), this);
    }
}

// @@@ SCORE_ENTRY* MONSTER::ScoreEntry()

SCORE_ENTRY* MONSTER::ScoreEntry()
{
    STRING name(LivingName());

    return new SCORE_ENTRY(MyScore, name, rip);
}

// @@@ comp MONSTER::compare(OBJECT* object)

comp MONSTER::compare(OBJECT* object)
{
    if(object->IsMonster(FORCE)) // Dead monsters before living ones!
    {
	MONSTER* monster = (MONSTER*) object;

	if(monster->dead() && !dead())
	    return GREATER_THAN;
	else if(dead() && !monster->dead())
	    return LESS_THAN;
    }
    comp c = OBJECT::compare(object);

    if(c == EQUAL)
	return ::compare(identity(), object->identity());
    else
	return c;
}

// @@@ void MONSTER::eat(LEVEL* level, MONSTER* monster)

void MONSTER::eat(LEVEL* level, MONSTER* monster)
{
    monster->pack.destroy(this);

    if(dead())
    {
	monster->message(format("You eat %s", ::print(this, PRINT_OBJECT)));
	level->WatchMonster(monster, "%s eats %s", this);
    }
    else
    {
	bug.error("MONSTER::eat", "Eating living monster...");
	monster->message("You must kill it first!");
    }
}

// @@@ void MONSTER::apply(LEVEL* level, MONSTER* monster)

void MONSTER::apply(LEVEL* level, MONSTER* monster)
{
    eat(level, monster);

    if(dead())
	body->apply(level, monster);
}

// @@@ QUESTION* MONSTER::DoApply(LEVEL* level

QUESTION* MONSTER::DoApply(LEVEL* level, PLAYER* player, CHANNEL* channel)
{
    eat(level, player);

    if(dead())
	return body->DoApply(level, player, channel);
    else
	return NULL;
}

// @@@ void MONSTER::SpendTime(LEVEL* level)

void MONSTER::SpendTime(LEVEL* level)
{
    // Update body && pack
    if(pack.check())
	pack.update(level, where, this);
    body->update(level, this);
    
    // Teleport at random
    int val = GetValue(ABILITY_RANDOM_TELEPORT);
    if(val > 0 && rnd.chance(val, 100))
	teleport(level, val);

    // Extra healing
    val = GetValue(ABILITY_HEALTH);
    if(val > 0 && body->wounded() && rnd.chance(val, 100))
	body->heal(val / 5 + 1);
    
    // Extra digestion
    val = GetValue(ABILITY_HUNGER);
    if(val > 0)
        body->DigestFood(this, val);
}

// @@@ cont MONSTER::play(LEVEL*)

cont MONSTER::play(LEVEL*)
{
    return DONE;
}

// @@@ bool MONSTER::AcceptEnter(MONSTER)

bool MONSTER::AcceptEnter(MONSTER*)
{ return dead(); }

// @@@ void MONSTER::defend(MONSTER* monster, LEVEL* level)
//
// Fight any monsters who attacks.

cont MONSTER::defend(MONSTER* monster, LEVEL* level)
{
    if(dead())			// Dead monsters do not defend themself...
	return CONTINUE;
    
    level->fight(monster, this);
    return DONE;
}

// @@@ bool MONSTER::catch(int power, OBJECT* object)

bool MONSTER::catch(int power, OBJECT* object)
{
    if(dead())
	return FALSE;

    if(rnd(experience() * knowledge(object) +
	   MyLog(object->size()) / 2) >
       rnd(power * 25 + MyLog(object->weight())))
    {
	pack.add(object);
	return TRUE;
    }
    return FALSE;
}

// @@@ bool MONSTER::hit(int power, OBJECT* weapon, bMONSTER* attack)

bool MONSTER::hit(int power, OBJECT* weapon, MONSTER* attack)
{
    if(dead())
	return OBJECT::hit(power, weapon, attack);

    // Yes, we do hit
    int Yes = MyLog(power);

    Yes += attack->experience();
    Yes += attack->dexterity();
    Yes += attack->knowledge(this) / 5;
    Yes += attack->knowledge(weapon) / 10;
    Yes += weapon->HitBonus(attack, this);
    Yes += MyLog(attack->speed());
    Yes += MyLog(this->size());
    Yes += ((this->pack.weight() * 10) / this->body->carry());

    if(this->sleeping())
	Yes += 50;
    if(this->IsConfused())
	Yes += 10;
    if(!attack->visible() && !this->SeeInvisible())
	Yes += 20;
    
    // No, we do not hit
    int No = 0;

    No += this->experience();
    No += this->dexterity();
    No += this->knowledge(attack) / 5;
    No += MyLog(this->speed());
    No += ((attack->pack.weight() * 10) / attack->body->carry());

    if(attack->IsConfused())
	No += 10;
    if(!this->visible() && !attack->SeeInvisible())
	No += 20;
    if(this->CheckAbility(ABILITY_PHASING))
	No += 50;
    
    return rnd(0, Yes) > rnd(0, No);
}

// @@@ STRING MONSTER::hurt(int power,

STRING MONSTER::hurt(int power, OBJECT* weapon, MONSTER* attack, LEVEL* level)
{
    if(dead())
	return OBJECT::hurt(power, weapon, attack, level);

    damage_type type = pack.absorb(&power, weapon, attack, this);

    if((type & DAMAGE_DRAIN_LIFE) && rnd.chance(1, 3))
    {
	message("You fell your life drain away!");
	GainExperience(- attack->experience());
    }

    if(body->hurt(power, type, weapon, attack, this, level) == DONE)
    {
	// You learn more from killing monsters smarter than you
	long exp = max(1L, 2L * experience() - attack->experience());
	// And smart monsters is always more rewarding than stupid monsters
	exp *= experience();
	// But you do not learn as much next time you kill it
	if(attack->IsPlayer())
	    exp *= 25L - attack->knowledge(this);
	else
	    exp *= 20L;
	
	if(exp > 0)		// You newer get *less* experienced...
	    attack->GainExperience(exp);
	
	if(attack == this)	// Do not gain anything from suicide!
	    kill("commit suicide", level);
	else
	{
	    attack->MyScore += exp; // Killing affect your score
	    attack->learn(this);    // Learn about what you just killed...
	    kill(format("was killed by %s's %s",
			::print(attack, PRINT_TYPE | PRINT_COUNT),
			::print(weapon, PRINT_NAME)), level);
	}
	return STRONG("The last blow did the trick...");
    }
    return NullString;
}

// @@@ int MONSTER::strength()

int MONSTER::strength()
{
    long tmp = (body->carry() - pack.weight()) * body->strength();
    tmp /= body->carry();
    
    return max(1, (int) tmp + pack.GetValue(ABILITY_STRENGTH));
}

// @@@ bool MONSTER::CanMove(long extra = 0L)

bool MONSTER::CanMove(long extra = 0L)
{
    return body->carry() > pack.weight() + extra && body->CanMove(this); 
}

// @@@ bool MONSTER::HeavyLoad(long extra = 0L)

bool MONSTER::HeavyLoad(long extra = 0L)
{
    return body->carry() * 2 < (pack.weight() + extra) * 3;
}

// @@@ bool MONSTER::LightLoad(long extra = 0L)

bool MONSTER::LightLoad(long extra = 0L)
{
    return body->carry() > (pack.weight() + extra) * 2;
}

// @@@ bool MONSTER::CanFight()

bool MONSTER::CanFight()
{
    return body->CanFight(this);
}

// @@@ void MONSTER::GainExperience(int exp)

void MONSTER::GainExperience(int exp)
{
    int OldExperience = MyExperience;

    MyExpPoint += exp;
    
    if(MyExpPoint < 0)
    {
	message("Hey, what going on?");
	MyExperience = 0;
	MyExpPoint = 0L;
    }
    else if(MyExpPoint >= (2 << MyExperience))
    {
	// Do not accept more than extr experience level
	MyExpPoint = (2 << MyExperience);
	MyExperience++;
    }
    else
    {
	while(MyExperience > 1 && MyExpPoint < (1 << MyExperience))
	    MyExperience--;
    }
    if(OldExperience != MyExperience)
    {
	// Make the monster healthier
	body->SetExperience(MyExperience);
	// Tell the player
	SendStatus(STATUS_EXPERIENCE, MyExperience);
    }
}

// @@@ void MONSTER::SetExperience(int exp)

void MONSTER::SetExperience(int exp)
{
    MyExperience = max(1, min(25, exp));
    MyExpPoint = (1 << MyExperience);
    body->SetExperience(MyExperience);
    SendStatus(STATUS_EXPERIENCE, MyExperience);
}

// @@@ void MONSTER::sleep(LEVEL*, long l)

void MONSTER::sleep(LEVEL*, long l)
{ UseMovePoint((int) l); }

// @@@ OBJECT* MONSTER::polymorph(LEVEL* level)

OBJECT* MONSTER::polymorph(LEVEL* level)
{
    if(dead())
    {
	body->life();
	level->AddMonster(this);
    }
    else if(CheckAbility(ABILITY_YENDOR))
	message("The magic of yendor protect your disguise.");
    else
    {
	BUNCH* bunch = level->create(position());
	bunch->touch();
	delete body;
	body = CreateBody(max(level->experience(), experience()));
    }
    return NULL;
}

// @@@ void MONSTER::create()

void MONSTER::create()
{
    rip = STRONG("found a bug");
    body = NULL;
    MyLetter = 'a';
    MyExperience = 1;
    MyExpPoint = MyScore = 0L;
    MovePoint = FightPoint = 1;
    where = NoWhere;
    DungeonLevel = LEVEL_ERROR;
    NewLocation = '*';
}

// @@@ void MONSTER::save(BINARY* binary)

void MONSTER::save(BINARY* binary)
{
    OBJECT::save(binary);
    body->save(binary);
    binary->PutChar(MyLetter);
    binary->PutChar(MyExperience);
    binary->PutLong(MyExpPoint);
    binary->PutLong(MyScore);
    binary->PutShort(MovePoint);
    binary->PutShort(FightPoint);
    binary->PutCoor(where);
    pack.save(binary);
    binary->PutShort(DungeonLevel);
    binary->PutShort(NewLocation);
}

// @@@ MONSTER::MONSTER(BINARY* binary)

MONSTER::MONSTER(BINARY* binary) : (binary)
{
    body = LoadBody(binary);
    MyLetter = binary->GetChar();
    MyExperience = binary->GetChar();
    MyExpPoint = binary->GetLong();
    MyScore = binary->GetLong();
    MovePoint = binary->GetShort();
    FightPoint = binary->GetShort();
    where = binary->GetCoor();
    pack.load(binary);
    DungeonLevel = binary->GetShort();
    NewLocation = binary->GetShort();
}

// @@@ MONSTER::MONSTER(PARSE* parse, coor pos)

MONSTER::MONSTER(PARSE* parse, coor pos) : (parse)
{
    create();
    where = pos;
    
    for(;;)
    {
	const char* string = parse->GetWord();

	/* BODY */
	if(strcmp(string, "INVENTORY") == EQUAL)
	{
	    parse->skip(INPUT_COMMENT);
	    OBJECT_LIST list(parse, NoWhere);
	    pack.add(&list);
	}
	else if(strcmp(string, "EXPERIENCE") == EQUAL)
	{
	    parse->skip(INPUT_SPACE);
	    MyExperience = (char) parse->GetLong();
	    parse->skip(INPUT_COMMENT);
	}
	else if(strcmp(string, "BREAK") == EQUAL)
	    break;
	else
	{
	    parse->wrong("MONSTER::MONSTER",
			 format("Unknown field: %s", string));
	    break;
	}
    }
    parse->skip(INPUT_COMMENT);
}

// @@@ MONSTER::MONSTER()

MONSTER::MONSTER()
{ create(); }

// @@@ MONSTER::~MONSTER()

MONSTER::~MONSTER()
{
    if(body != NULL)
	delete body;
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
